//! Lock Token module
use borsh::{BorshDeserialize, BorshSerialize};
use chrono::{DateTime, Duration, Utc};
use hmac::{Hmac, Mac};
use rand::Rng;
use serde::{Deserialize, Serialize};
use sha2::Sha256;

type HmacSha256 = Hmac<Sha256>;

#[derive(thiserror::Error, Debug, Serialize)]
/// Error type
pub enum Error {
    /// Missing update token
    #[error("Missing update token")]
    MissingUpdateToken,

    /// Invalid signature
    #[error("Invalid signature")]
    InvalidSignature,

    /// I/O error
    #[error("IO error: {0}")]
    #[serde(serialize_with = "crate::serialize_error_to_string")]
    IO(#[from] std::io::Error),
}

#[derive(Debug)]
/// Lock Token Manager
pub struct TokenManager(pub Vec<u8>);

impl Default for TokenManager {
    fn default() -> Self {
        let mut rng = rand::thread_rng();
        let mut payload = [0u8; 10];
        rng.fill(&mut payload);
        Self(payload.to_vec())
    }
}

impl TokenManager {
    /// Checks if the given token is valid and still not expired
    pub fn verify(&self, token: Token, update_token: &Option<Vec<u8>>) -> Result<(), Error> {
        if !token.is_valid() {
            return Ok(());
        }

        let update_token = if let Some(update_token) = update_token {
            update_token
        } else {
            return Err(Error::MissingUpdateToken);
        };

        let mut mac = HmacSha256::new_from_slice(&self.0).expect("HMAC can take key of any size");
        mac.update(&token.to_bytes().unwrap_or_default());
        let result = mac.finalize().into_bytes();
        if &result[..] != update_token {
            Err(Error::InvalidSignature)
        } else {
            Ok(())
        }
    }

    /// Creates a new instance of the token
    pub fn new_token(&self, duration: Duration) -> (Token, Vec<u8>) {
        let mut rng = rand::thread_rng();
        let mut payload = [0u8; 10];
        rng.fill(&mut payload);

        let token = Token {
            owner: "owner".to_string(),
            payload: payload.to_vec(),
            expires_at: Utc::now() + duration,
        };
        let token_bytes = token.to_bytes().unwrap_or_default();
        let mut mac = HmacSha256::new_from_slice(&self.0).expect("HMAC can take key of any size");
        mac.update(&token_bytes);
        let proof = mac.finalize().into_bytes();

        (token, proof.to_vec())
    }
}

#[derive(Serialize, Deserialize, BorshSerialize, BorshDeserialize, Clone, Debug, PartialEq)]
/// Inner token
///
/// This token has the transaction ID, the expiration date and the owner string. This is the data to
/// be signed by the HMAC and sent to the client.
pub struct Token {
    #[borsh(
        serialize_with = "crate::to_ts_microseconds",
        deserialize_with = "crate::from_ts_microseconds"
    )]
    expires_at: DateTime<Utc>,
    payload: Vec<u8>,
    owner: String,
}

impl Token {
    /// Converts the token to bytes
    pub fn to_bytes(&self) -> Result<Vec<u8>, Error> {
        Ok(borsh::to_vec(&self)?)
    }

    /// Checks if the token is valid
    pub fn is_valid(&self) -> bool {
        self.expires_at > Utc::now()
    }
}
